#!/usr/bin/env ruby
require "date"
class DateTime
  alias __to_s to_s
  def to_s
    __to_s.sub(/Z$/,'')
  end
end

begin # for RubyGems users
  require "rubygems"
  require_gem "activerecord"
rescue LoadError
  # for Debian users
  $LOAD_PATH << "/usr/share/rails/activerecord/lib/"
  require "active_record"
end
require "kconv"
require "find"
require "optparse"
require "yaml"
require "numru/gphys"

CONFIG_PATH = File.join(File.dirname(__FILE__), '../config')
require File.join(CONFIG_PATH, "environment")

app_path = File.join(File.dirname(__FILE__), '../app')
require File.join(app_path, "models/directory")
require File.join(app_path, "models/keyword_attribute")
require File.join(app_path, "models/variable")
require File.join(app_path, "models/spatial_attribute")
require File.join(app_path, "models/actual_file")

CONFIG_DATABASE = YAML.load_file(File.join(CONFIG_PATH, "database.yml"))
DEFAULT_DATABASE = 'development'

GFDNAVI_CHARSET = Kconv::UTF8
GFDNAVI_LANGUAGES = ['ja', 'en']

module RegisterSharedTools
  #
  #hash ɓo^L[[hƒl̃Zbg^, entity_id ɂ
  #KeywordAttribute  belong_to ̊֌WɂfBNg܂
  #ϐ ID ^.
  #
  #L[[hƒl 2 oCg܂܂ꍇ, GFDNAVI_CHARSET
  #Őݒ肳镶R[hɕϊēo^.
  #
  # lang ݒ肳Ăꍇ,  lang ɐݒ肳镶
  #GFDNAVI_LANGUAGES ɓo^Ă̂Ȃ, ̕
  #keyword_attribute e[u lang ɓo^.
  #
  #lang  nil łꍇɂ, L[[h̖ ".ja" ̂悤
  #񂪂ꍇ, ̌ݒ肪 keyword_attribute e[u
  #lang ɓo^. ".ja" ̃TtBbNX͎菜.
  #
  #lang  nil ł, L[[h̖ ".ja" ̂悤Ȍ
  #TtBbNXȂꍇł, l 2 oCgłꍇɂ
  #"ja"  keyword_attribute e[u lang ɓo^.
  #
  def keyword_attribute_register(hash, dir_id=nil, var_id=nil, lang=nil)
    return if !(dir_id) && !(var_id)
    hash.each{ |key, value|
      keyatt = KeywordAttribute.new
      keyatt.directory_id = dir_id if dir_id
      keyatt.variable_id = var_id if var_id
      keyatt.lang = lang if lang && GFDNAVI_LANGUAGES.include?(lang)
      suffix = nil
      if !(keyatt.lang) && key =~ /\-(\w+)$/
        suffix = $1
        if GFDNAVI_LANGUAGES.include?(suffix)
          keyatt.lang = suffix
          key = key.sub(/\-\w+$/, '')
        end
      end
      if key.class.to_s == "String"
        keyatt.name = Kconv.kconv(key, GFDNAVI_CHARSET)
        if !(keyatt.lang) && ja_charset?(key)
          keyatt.lang = 'ja'
        end
      else
        keyatt.name = key
      end
      if value.class.to_s == "String"
        keyatt.value = Kconv.kconv(value, GFDNAVI_CHARSET)
        if !(keyatt.lang) && ja_charset?(value)
          keyatt.lang = 'ja'
        end
      elsif value.class.to_s == "NilClass"
        keyatt.value = ''
      elsif value.class.to_s == "NArray"
        keyatt.value = value
      else
        begin
          keyatt.value = NArray.to_na(value)
        rescue
          keyatt.value = value.to_s
        end
      end
      keyatt.save
    }
  end

  #
  #str ɗ^镶񂪓{ł true Ԃ.
  #
  def ja_charset?(str)
    guessed_code = Kconv.guess(str)
    if [Kconv::JIS, Kconv::EUC, Kconv::SJIS,
        Kconv::UTF8, Kconv::UTF16].include?(guessed_code)
      return true
    else
      return false
    end
  end

  module_function :keyword_attribute_register
  module_function :ja_charset?
end

#
#fBNgȉ̃fBNg񂨂т̃fBNgɕt
#SIGEN t@C gfdnavi f[^x[Xɓo^邽߂̃NX.
#
class RegisterDirTree

  # [gfBNg̖
  ROOTDIR_NAME = '/'
  ROOTPATH_NAME = '/'

  # fBNgƕϐ̋؂蕶
  DIR_VAR_DELIMITER = Variable::DIR_VAR_DELIMITER

  # YAML t@C, GPhys ϐɏt邽߂̃L[
  YAML_GPHYSVAR_KEYWORD = 'contains'

  #
  #\bh.
  #
  #topdir ɂ̓gbvfBNg^. ftHg,
  #config/environment.rb Ŏw肳郋[gfBNgł.
  #
  #database ɂ 'development', 'production', 'test' ̂ꂩ
  #^. ftHg 'development' ł.
  #
  #quiet  true ^, i󋵂̏o͂sȂ.
  #ftHg͐i󋵂̏o͂s.
  #
  def initialize(topdir=GFDNAVI_DATA_PATH, database=nil, quiet=nil)
    topdir ||= GFDNAVI_DATA_PATH
    raise IOError, "  Error: \"#{topdir}\" is not found.\n" if !(File.exist?(topdir))
    raise IOError, "  Error: \"#{topdir}\" is not a directory.\n" if !(File.directory?(topdir))
    @topdir = File.expand_path(topdir)
    @quiet = quiet
    verbose_msg "\"#{topdir}\" is set as top directory.\n"

    @database = database || ENV['RAILS_ENV'] || DEFAULT_DATABASE

    ActiveRecord::Base.establish_connection(
                       :adapter => CONFIG_DATABASE[@database]['adapter'],
                       :host => CONFIG_DATABASE[@database]['host'],
                       :username => CONFIG_DATABASE[@database]['username'],
                       :password => CONFIG_DATABASE[@database]['password'],
                       :database => CONFIG_DATABASE[@database]['database']
                                            )
  end

  #
  #̃\bhĂяo, i󋵂̏o͂sȂ悤ɂȂ.
  # false ^, i󋵂̏o͂s悤ɂȂ.
  #
  def quiet(flag=true)
    @quiet = flag
  end

  #
  #dirname ɗ^fBNgȉ̃fBNg, SIGEN t@C
  #͂, f[^x[Xɓo^.
  #parent_id ɂ dirname ̃f[^x[X ID ^.
  #dirname ܂ parent_id ȗꍇɂ, new \bh
  #^ topdir fBNgƂ.
  #
  def register(dirname=nil, parent_id=nil)
    if !(parent_id) || !(dirname)
      topdir = Directory.new
      topdir.name = ROOTDIR_NAME
      topdir.path = ROOTPATH_NAME
      topdir.save
      verbose_msg "Insert info from \"... /\".\n"
      parent_id = topdir.id
      dirname = @topdir
    end
    begin
      Dir.chdir(dirname)
    rescue
      verbose_msg "Skip \"... #{dirname.sub(/^#{@topdir}/, '')}\" for \"#{$!.to_s}\"\n"
      return false
    end
    filelist = Dir.glob("*")
    filelist.each{ |file|
      if !File.readable?(file) || \
        File.directory?(file) && !(/mswin/=~Config::CONFIG["arch"]) && !File.executable?(file)
        verbose_msg "Skip \"... #{File.join(dirname.sub(/^#{@topdir}/, ''), file)}\" for \"Permission denied\"\n"
        next
      end

      ignored_flag = false
      if GFDNAVI_IGNORED_DIRS
        GFDNAVI_IGNORED_DIRS.each{|igdir|
          if igdir.class.to_s == "String"
            ignored_flag = true if igdir == file
          elsif igdir.class.to_s == "Regexp"
            ignored_flag = true if igdir =~ file
          end
        }
      end
      next if ignored_flag
      if File.directory?(file)
        dir = Directory.new
        dir.name = file
        dir.path = File.join(dirname, file).sub(/^#{@topdir}/, '')
        dir.parent_id = parent_id
        dir.save
        verbose_msg "Insert info from \"... #{dir.path}\".\n"
        meta_files_register(file, dir.id)
        # recursive search
        register(File.join(dirname, file), dir.id)
      elsif NumRu::GPhys::IO.file2type(file)
        gphys_reg = GPhysRegister.new(file, parent_id, @topdir)
        next if gphys_reg.error
        gphys_reg.quiet if @quiet
        dir_id = gphys_reg.register
        gphys_reg.close
        meta_files_register(file, dir_id)
      end
    }
    Dir.chdir("..") if parent_id && dirname
  end

  #
  #^f[^i[t@Cǂݍ, fBNg
  #̂߂f[^Ƃēo^. , ȉ̎ނ̃t@C
  #ǂݍ.
  #
  def meta_files_register(file, entity_id)
    yaml_register(file+".yml", entity_id)
    GFDNAVI_LANGUAGES.each{ |lang|
      yaml_register(file+".yml."+lang, entity_id, lang)
    }
    sigen_register(file+".SIGEN", entity_id)
    GFDNAVI_LANGUAGES.each{ |lang|
      sigen_register(file+".SIGEN."+lang, entity_id, lang)
    }
  end

  #
  #yamlfile  YAML t@C^. entity_id ɂ YAML 
  #wfBNgt@C^. yamlfile 
  #Ȃꍇ͉Ȃł̃\bhI, e\bhɖ߂.
  #
  def yaml_register(yamlfile, entity_id, lang=nil)
    return unless File.file?(yamlfile)
    path = File.expand_path(yamlfile).sub(/^#{@topdir}/, '')
    begin
      yaml_info = YAML.load_file(yamlfile)
    rescue
      verbose_msg "Skip \"... #{path}\" for \"#{$!.to_s}\"\n"
      return false
    end
    yaml_gphysvar_info = yaml_info.delete(YAML_GPHYSVAR_KEYWORD)
    RegisterSharedTools::keyword_attribute_register(yaml_info, entity_id, nil, lang)

    yaml_gphysvar_info.each{|varname, info|
      if varname =~ /(.*)[#{DIR_VAR_DELIMITER}\/](\w+)( +in +(\S+))?$/
        yaml_gphys_filename = $1
        yaml_gphysvar_name = $2
        yaml_gphys_alias = $4    # nil if ' as *' is not given (horinout)
        yaml_gphysdir_path = 
                   File.join(path.sub(/\.yml$/, '').sub(/.*\//, ''),
                        yaml_gphys_filename)
        yaml_gphys_files = Dir.glob(yaml_gphysdir_path).sort
	if yaml_gphys_alias
	  entity_id = nil   # directoryύX悤eID͕sƂ(horinout)
	  yaml_gphysdir_path_in_db = 
                 File.expand_path(  \
                       File.join(path.sub(/\.yml$/, '').sub(/.*\//, ''), \
			         yaml_gphys_alias) , '/'    \
                 ).sub(/^#{File.expand_path(".","\/")}/,'')
	else
	  yaml_gphysdir_path_in_db = yaml_gphysdir_path
	end 

        yaml_gphysdir = 
          GPhysRegister.new(yaml_gphys_files,
                            entity_id,
                            @topdir,
                            yaml_gphysdir_path_in_db,
                            [yaml_gphysvar_name])

        next if yaml_gphysdir.error
        yaml_gphysdir.quiet if @quiet
        yaml_gphysdir_id = yaml_gphysdir.register
        yaml_gphysdir.close

        RegisterSharedTools::keyword_attribute_register(info, yaml_gphysdir_id, nil, lang)

      else
        yaml_gphysvar_path = path.sub(/\.yml$/, '') +
          DIR_VAR_DELIMITER + varname
        yaml_gphysvar_list =
          Variable.find_by_sql(["select * from variables where path like ? order by path",
                                 yaml_gphysvar_path])
        yaml_gphysvar_list.each{|yaml_gphysvar|
          RegisterSharedTools::keyword_attribute_register(info, nil, yaml_gphysvar.id, lang)
        }
      end
    } if yaml_gphysvar_info
    verbose_msg "Insert info from \"... #{path}\".\n"
  end


  #
  #sigenfile  SIGEN t@C^. entity_id ɂ SIGEN 
  #wfBNgt@C^. sigenfile 
  #Ȃꍇ͉Ȃł̃\bhI, e\bhɖ߂.
  #
  def sigen_register(sigenfile, entity_id, lang=nil)
    return unless File.file?(sigenfile)
    path = File.expand_path(sigenfile).sub(/^#{@topdir}/, '')
    sigen = Sigen.new(sigenfile)
    sigen_info = sigen.parse
    RegisterSharedTools::keyword_attribute_register(sigen_info, entity_id, nil, lang)
    verbose_msg "Insert info from \"... #{path}\".\n"
  end

  def verbose_msg(msg)
    return if @quiet
    print "  #{self.class}: #{msg}"
  end

  #
  #GPhys[http://www.gfd-dennou.org/library/ruby/]
  #ňƂ\ȃf[^f[^x[XɊi[.
  #
  class GPhysRegister
    include NumRu

    attr_reader :error

    #
    #\bh. *file* ɂ String NX̃t@C, 
    #܂̓t@Ci[ Array NX^. Array NX
    #Ńt@C^ꍇ, X̃t@CʌɃf[^x[X
    #o^̂ł͂Ȃ, Array NXɋLqSẴt@C
    #zIȃt@Cf[^x[XւƓo^.
    #̍, zIȃt@ĆuĂіv *united_name*
    #ɗ^Kv (: "*.nc"). *united_name* ^Ȃ
    #ꍇ, Array NX̐擪̃t@Ĉ݂f[^x[Xւ
    #o^.
    #
    # *parent_id* ɂ *file* ̐efBNg
    #f[^x[X ID ^, *topdir*
    #ɂ̓f[^x[X̃gbvfBNg^.
    #
    # *varnames* ͕ϐ Array NXɊi[ė^.
    #̈^ꂽꍇ, ̕ϐ݂̂f[^x[Xւ
    #i[.
    #
    def initialize(file, parent_id, topdir, united_name=nil, varnames=nil)
      @file_list = []
      @united_flag = false
      @varnames = nil
      if String === file
        @file_list << file
        @filename = @file_list[0]
      elsif Array === file
        if String === united_name
          @file_list = file
          @filename = united_name
          @united_flag = true
        else
          @file_list << file[0]
          @filename = @file_list[0]
        end
      else
        raise ArgumentError, " Error: line __LINE__ in __FILE__, \"file\" argument is " +
          " \"#{file.class.to_s} class. This class is invalid.\n"
      end
      if Array === varnames
        @varnames = []
        varnames.each{|v| @varnames << v.to_s}
      end

      @topdir_path = File.expand_path(topdir)
      @relative_path = File.expand_path(@filename).sub(/^#{@topdir_path}/, '')
      if parent_id
	@parent_id = parent_id
      else
	@parent_id = Directory.find_by_path(File.dirname(@relative_path)).id
      end
      @info = []
      @error = true
      @varlist_except_coordinates = nil
      @file_obj_list = []
      @gphysdir_file_list = []
      @time_info = []
      @dir_longitude_info = []
      @dir_latitude_info = []

      @file_list.each{|file|
        file_path = File.expand_path(file)
        raise IOError, "  Error: \"#{@topdir_path}\" is not found.\n" if !(File.exist?(@topdir_path))
        raise IOError, "  Error: \"#{@topdir_path}\" is not a directory.\n" if !(File.directory?(@topdir_path))
        raise ArgumentError, "  Error: \"parent_id\" is invalid.\n" if !(@parent_id)
        if !(file_path =~ /^#{@topdir_path}/)
          raise ArgumentError, "  Error: \"#{file_path}\" is not included in \"#{@topdir_path}\".\n"
        end
        rel_file_path = file_path.sub(/^#{@topdir_path}/, '')
        begin
          @file_obj_list << GPhys::IO.file2file_class(file_path).open(file_path)
          @error = false
        rescue
          verbose_msg "Skip \"... #{rel_file_path}\" for \"#{$!.to_s}\"\n"
          next
        end
        varlist_except_coordinates_tmp = GPhys::IO.var_names_except_coordinates(@file_obj_list[-1])
        if !(@varlist_except_coordinates)
          @varlist_except_coordinates = varlist_except_coordinates_tmp
        else
          @varlist_except_coordinates.collect!{|checkv|
            if varlist_except_coordinates_tmp.include?(checkv)
              checkv
            else
              nil
            end
          }
          @varlist_except_coordinates.compact!
        end
        varlist = GPhys::IO.var_names(@file_obj_list[-1])
        gphysdir = GPhys::IO.open(@file_obj_list[-1], varlist[0])
        @gphysdir_file_list << gphysdir.data.file
        @time_info = self.parse_time(@file_obj_list[-1], varlist, @time_info)
        @dir_longitude_info = self.parse_longitude(@file_obj_list[-1], varlist, @dir_longitude_info)
        @dir_latitude_info = self.parse_latitude(@file_obj_list[-1], varlist, @dir_latitude_info)
      }

      if @varnames
        @varlist_except_coordinates.collect!{|checkv|
          if @varnames.include?(checkv)
            checkv
          else
            nil
          end
        }
        @varlist_except_coordinates.compact!
      end

    end

    #
    #I\bh. open Ă GPhys ϐ.
    #(GPhys ϐĴ 30 ܂łƂ炵.)
    #
    def close
      @file_obj_list.each{|o| o.close}
    end

    #
    #̃\bhĂяo, i󋵂̏o͂sȂ悤ɂȂ.
    # false ^, i󋵂̏o͂s悤ɂȂ.
    #
    def quiet(flag=true)
      @quiet = flag
    end

    #
    #new \bh *file* ɗ^f[^̏f[^x[X
    #i[.
    #
    #Ԃl͓o^fBNg ID ł.
    #
    def register
      dir_id = self.register_directory
      @varlist_except_coordinates.each{|varname|
        if @file_obj_list.size < 2
          file_obj = @file_obj_list[0]
        else
          file_obj = @file_obj_list
        end
        gphysvar = GPhysVariableRegister.new(file_obj, varname, dir_id, @relative_path)
        gphysvar.quiet if @quiet
        gphysvar.time_var(@time_info)
        gphysvar.longitude_var(@dir_longitude_info)
        gphysvar.latitude_var(@dir_latitude_info)
        varid = gphysvar.register
	register_actual_file(varid) if @file_list.length >=2
      }
      dir_id
    end

    def register_actual_file(varid)
      @file_list.each do |file|
	af = ActualFile.new
        af.path = File.expand_path(file).sub(/^#{@topdir_path}/, '')
	af.variable_id = varid
	af.save
      end
    end

    #
    #new \bh *file* ɗ^f[^ɂ, eϐł͂Ȃ,
    #f[^Ŝ̏f[^x[XɊi[. Ԃl̓f[^x[X
    #̃fBNg ID ł.
    #
    def register_directory
      dir = Directory.new
      dir.name = File.basename(@filename)
      dir.path = @relative_path
      dir.parent_id = @parent_id
      dir.plain_file = true
      dir.save

      dir_info = self.parse
      RegisterSharedTools::keyword_attribute_register(dir_info, dir.id)

      if @dir_longitude_info || @dir_latitude_info
        spatt = SpatialAttribute.new
        spatt.directory_id = dir.id
        if @dir_longitude_info
          spatt.longitude_lb = @dir_longitude_info[0]['longitude_lb']
          spatt.longitude_rt = @dir_longitude_info[0]['longitude_rt']
          spatt.lon_ax_name = @dir_longitude_info[0]['name']
        end
        if @dir_latitude_info
          spatt.latitude_lb = @dir_latitude_info[0]['latitude_lb']
          spatt.latitude_rt = @dir_latitude_info[0]['latitude_rt']
          spatt.lat_ax_name = @dir_latitude_info[0]['name']
        end
        spatt.save
      end

      verbose_msg "Insert info from \"... #{dir.path}\".\n"
      dir.id
    end

    #
    #GPhys fBNg̓\bh. ͂e Hash NX
    #i[ĕԂ.
    #
    def parse
      @info = {}
      @gphysdir_file_list.each{|gphys_file|
        gphys_file.each_att{|att|
          if !(@info[att.name])
            @info[att.name] = att.get
          end
        }
      }
      @info
    end

    #
    #== WɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name      :: ϐ̖O (String NX)
    #starttime :: Jn       (DateTime NX)
    #enddtime  :: I       (DateTime NX)
    #
    #O *prev_info* ɂ͊ parse_time 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #starttime ͂菬Ȓl, endtime ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== W̔@
    #
    #ȉ̍WoxWƔF
    #
    #* units  NumRu::Units#reduce5 ɂĕϊ̂
    #  "86400 s @ 2002-01-01T00:00:00.00 +00:00" ̂悤ȃtH[}bg
    #  ƂȂĂ邱. NumRu::Units#reduce5 ɂĂ
    #  {NumRu::Units}[http://ruby.gfd-dennou.org/products/numru-units/doc/units.html]
    #  QƂ̂.
    #
    def parse_time(file_obj, varlist, prev_info=[])
      time_info = prev_info
      varlist.each{|varname|
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        gphysvar.att_names.each{|attname|
          if attname == 'units'
            units = Units.new(gphysvar.get_att(attname))
            if units.reduce5.to_s =~ /^\((\d+ )?s @ (\d{4}\-\d{2}\-\d{2}T\d{2}:\d{2}:\d{2}\.\d{2} [\+\-]\d{2}:\d{2})\)$/
              info['name'] = varname
              unit_no_offset = $1 || 1
              unit_no_offset = unit_no_offset.to_f
              offset = DateTime.parse($2)
              start_val = gphysvar.val[0]
              end_val = gphysvar.val[-1]
              info['starttime'] = offset + (start_val * (unit_no_offset / 86400.0))
              info['endtime'] = offset + (end_val * (unit_no_offset / 86400.0))
              time_info.collect!{|prev|
                if prev['name'] == info['name']
                  if prev['starttime'] < info['starttime']
                    info['starttime'] = prev['starttime']
                  end
                  if prev['endtime'] > info['endtime']
                    info['endtime'] = prev['endtime']
                  end
                  nil
                else
                  prev
                end
              }
              time_info.compact!
              time_info << info
            end
          end
        }
      }
      if time_info.size > 0
        time_info
      else
        nil
      end
    end

    #
    #== oxWɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #ox̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name         :: oxϐ̖O (String NX)
    #longitude_lb :: ox̐[     (Float NX)
    #longitude_rt :: ox̓[     (Float NX)
    #
    #O *prev_info* ɂ͊ parse_longitude 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #longitude_lb ͂菬Ȓl, longitude_rt ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== oxW̔@
    #
    #ȉ̍WoxWƔF
    #
    #* ϐ *lon* ܂ *longitude* ł邱 (啶͋ʂȂ)
    #* units  *degree_east* ܂ *degrees_east* ł邱.
    #
    #
    def parse_longitude(file_obj, varlist, prev_info=[])
      lon_info = prev_info
      varlist.each{|varname|
        flag = nil
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        if varname =~ /^lon(gitude)?$/i
          flag = true
        else
          gphysvar.att_names.each{|attname|
            if attname == 'units'
              if gphysvar.get_att(attname) =~ /^degrees?_east$/
                flag = true
              end
            end
          }
        end
        if flag
          info['name'] = varname
          info['longitude_lb'] =  gphysvar.val[0]
          info['longitude_rt'] = gphysvar.val[-1]
          lon_info.collect!{|prev|
            if prev['name'] == info['name']
              if prev['longitude_lb'] < info['longitude_lb']
                info['longitude_lb'] = prev['longitude_lb']
              end
              if prev['longitude_rt'] > info['longitude_rt']
                info['longitude_rt'] = prev['longitude_rt']
              end
              nil
            else
              prev
            end
          }
          lon_info.compact!
          lon_info << info
        end
      }
      if lon_info.size > 0
        lon_info
      else
        nil
      end
    end

    #
    #== ܓxWɊւT\bh
    #
    # *file_obj* ɗ^t@C܂ GPhys::IO
    #IuWFNg *varlist* ɗ^ϐ̒, 
    #ܓx̕ϐ̏Ԃ. Ԃ Array NXɂȂĂ.
    #Array ̌X̗vf Hash NX琬.
    #ꂼ̃L[̈ӖȉɋL.
    #
    #name        :: ܓxϐ̖O (String NX)
    #latitude_lb :: ܓx̓[     (Float NX)
    #latitude_rt :: ܓx̖k[     (Float NX)
    #
    #O *prev_info* ɂ͊ parse_latitude 
    #xꂽ^. prev_info Ɋi[Ă
    #̒, name ̂ɊւĂ,
    #latitude_lb ͂菬Ȓl, latitude_rt ͂傫Ȓl
    #Hash Ɋi[ĕԂ.
    #
    #=== ܓxW̔@
    #
    #ȉ̍WܓxWƔF
    #
    #* ϐ *lat* ܂ *latitude* ł邱 (啶͋ʂȂ)
    #* units  *degree_north* ܂ *degrees_north* ł邱.
    #
    #
    def parse_latitude(file_obj, varlist, prev_info=[])
      lat_info = prev_info
      varlist.each{|varname|
        flag = nil
        info = {}
        gphysvar = GPhys::IO.open(file_obj, varname)
        if varname =~ /^lat(itude)?$/i
          flag = true
        else
          gphysvar.att_names.each{|attname|
            if attname == 'units'
              if gphysvar.get_att(attname) =~ /^degrees?_north$/
                flag = true
              end
            end
          }
        end
        if flag
          info['name'] = varname
          if gphysvar.val[0] < gphysvar.val[-1]
            info['latitude_lb'] =  gphysvar.val[0]
            info['latitude_rt'] = gphysvar.val[-1]
          else
            info['latitude_lb'] = gphysvar.val[-1]
            info['latitude_rt'] =  gphysvar.val[0]
          end
          lat_info.collect!{|prev|
            if prev['name'] == info['name']
              if prev['latitude_lb'] < info['latitude_lb']
                info['latitude_lb'] = prev['latitude_lb']
              end
              if prev['latitude_rt'] > info['latitude_rt']
                info['latitude_rt'] = prev['latitude_rt']
              end
              nil
            else
              prev
            end
          }
          lat_info.compact!
          lat_info << info
        end
      }
      if lat_info.size > 0
        lat_info
      else
        nil
      end
    end

    def verbose_msg(msg)
      return if @quiet
      print "  #{self.class}: #{msg}"
    end

  end

  class GPhysVariableRegister
    include NumRu
    #
    #\bh. *file* ɂ̓t@C, *varname* ɂ
    #ϐ^.
    #
    def initialize(file, varname, directory_id, directory_path)
      @varname = varname
      @gphysvar = GPhys::IO.open(file, @varname)
      @directory_id = directory_id
      @path = directory_path + DIR_VAR_DELIMITER + @varname
      @time_info = nil
      @dir_longitude_info = nil
      @dir_latitude_info = nil
    end

    #
    #̃\bhĂяo, i󋵂̏o͂sȂ悤ɂȂ.
    # false ^, i󋵂̏o͂s悤ɂȂ.
    #
    def quiet(flag=true)
      @quiet = flag
    end

    #
    #hash ɎԕϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐԕϐɈˑꍇ, ^ꂽ񂩂 starttime, endtime
    #Ƀf[^o^.
    #
    def time_var(info)
      @time_info = info
    end

    #
    #hash ɌoxϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐoxϐɈˑꍇ, ^ꂽ, spatial_attribute
    #e[uɃf[^o^.
    #
    def longitude_var(info)
      @dir_longitude_info = info
    end

    #
    #hash ɈܓxϐɊւ^邱Ƃ,  new œo^ꂽ
    #ϐܓxϐɈˑꍇ, ^ꂽ, spatial_attribute
    #e[uɃf[^o^.
    #
    def latitude_var(info)
      @dir_latitude_info = info
    end

    #
    #new \bh *file*, *varname* ɗ^ϐ̏f[^x[X
    #i[. Ԃlɕϐ̃f[^x[X ID Ԃ.
    #
    def register
      var = Variable.new
      var.path = @path
      var.name = @varname
      var.directory_id = @directory_id
      if @time_info
        @time_info.each{|info|
          if @gphysvar.axnames.include?(info['name'])
            var.starttime = info['starttime']
            var.endtime = info['endtime']
            break
          end
        }
      end
      var.save

      var_info = self.parse_attributes
      RegisterSharedTools::keyword_attribute_register(var_info, nil, var.id)

#      if @dir_longitude_info || @dir_latitude_info
#        spatt = SpatialAttribute.new
#        spatt.variable_id = var.id
#        if @dir_longitude_info
#          @dir_longitude_info.each{|info|
#            if @gphysvar.axnames.include?(info['name'])
#              spatt.longitude_lb = info['longitude_lb']
#              spatt.longitude_rt = info['longitude_rt']
#              spatt.lon_ax_name = info['name']
#            end
#          }
#        end
#        if @dir_latitude_info
#          @dir_latitude_info.each{|info|
#            if @gphysvar.axnames.include?(info['name'])
#              spatt.latitude_lb = info['latitude_lb']
#              spatt.latitude_rt = info['latitude_rt']
#              spatt.lat_ax_name = info['name']
#            end
#          }
#        end
#        spatt.save

      if @dir_longitude_info || @dir_latitude_info
        spatt = SpatialAttribute.new
        spatt.variable_id = var.id
        if @dir_longitude_info
          @dir_longitude_info.each{|info|
            spatt.longitude_lb = info['longitude_lb']
            spatt.longitude_rt = info['longitude_rt']
            spatt.lon_ax_name = info['name']
          }
        end
        if @dir_latitude_info
          @dir_latitude_info.each{|info|
            spatt.latitude_lb = info['latitude_lb']
            spatt.latitude_rt = info['latitude_rt']
            spatt.lat_ax_name = info['name']
          }
        end
        spatt.save


      end

      verbose_msg "Insert info from \"... #{@path}\".\n"

      var.id
    end

    #
    #GPhys ϐ̑Ɋւ̓\bh. ͂e Hash NX
    #i[ĕԂ.
    #
    def parse_attributes
      @info = {}
      @gphysvar.att_names.each{|attname|
        @info[attname] = @gphysvar.get_att(attname)
      }
      @info
    end

    def verbose_msg(msg)
      return if @quiet
      print "  #{self.class}: #{msg}"
    end

  end



end


#
#fBNgf[^x[X}l[W
#mksigen[http://www.gfd-dennou.org/arch/cc-env/mksigen/SIGEN.htm]
# SIGEN t@C̉́E쐬pNX
#
class Sigen

  #
  #\bh. *file* ɂ̓t@C^.
  #
  #͂̏ꍇ͑ *rw* ^Kv͂Ȃ. *file* ݂Ȃ
  #ꍇ, G[Ԃ.
  #
  #쐬̏ꍇ͑ *rw*  "w"  "w+" ^.
  #
  def initialize(file, rw="r")
    @filename = File.expand_path(file)
    @io = open(@filename, rw)
    @rw = rw
    @info = {}
  end

  #
  #SIGEN t@C̓\bh. ͂e Hash ƂĕԂ.
  #Hash ̃L[ɃtB[h, lɂ̓tB[h{fBi[
  #. tB[h͏ɕϊ. tB[h{fB
  #̉s╡̋󔒕͑Sċ󔒈ꕶɕϊ.
  #tB[h, tB[h{fB̕R[h UTF-8 
  #ϊ.
  #
  def parse
    if @rw == "w"
      raise IOError, "  Error: #{@filename} was opened writable mode.\n"
    end
    name = ""
    @io.each{ |line|
      line.chomp!
      line = Kconv.kconv(line, GFDNAVI_CHARSET)
      break if line =~ /^\s*$/u
      if ! (line =~ /^\s/u)
        if ! (line =~ /^(\w+):\s*(.*)/u)
          warn "  Warning: broken header \"#{line}\", " +
            "line #{@io.lineno} in \"#{@filename}\"\n"
          name = ""
          next
        end
        name = $1.tr('A-Z', 'a-z')
        val = $2.gsub(/[\t\r]/u, ' ')
        if @info[name] then
          @info[name] << ' ' + val
        else
          @info[name] = val
        end
      else
        next if name.empty?
        @info[name] << line.gsub(/[\t\r]/u, ' ').gsub(/^\s*/u, ' ')
      end
    }
    @info
  end

  # SIGEN t@C쐬p\bh܂ĂȂ.
  def generate(sigen_hash)
    
  end
end



#
# ȉ͂̃t@CsvOƂēۂ̓.
#
if $0 == __FILE__
  opt = OptionParser.new
  OPTS = {}
  ARGV.options{|opt|
    opt.on( '-T=VAL', '--topdir=VAL',
            "Set top directory 
                                     (default: #{GFDNAVI_DATA_PATH})"
            ){|v| OPTS[:topdir] = v.gsub(/^=/, '')}

    opt.on( '-P', '--production',
            "Use database \"#{CONFIG_DATABASE['production']['database']}\""
            ){|v| OPTS[:database] = 'production'}

    opt.on(       '--test',
            "Use database \"#{CONFIG_DATABASE['test']['database']}\"
                                     (default: \"#{CONFIG_DATABASE[DEFAULT_DATABASE]['database']}\")"
            ){|v| OPTS[:database] = 'test'}

    opt.on( '-q', '--quiet',
            "Quiet mode"
            ){|v| OPTS[:quiet] = v}

    opt.on_tail('-h', '-H', '--help', 
                "This help message is output"
                ){|v| OPTS[:help] = v}
  opt.parse!
  }

  if OPTS[:help]
    print <<-"EOF"

  #{File.basename($0.to_s)}:

    USAGE: #{File.basename($0.to_s)} [options]

    OPTION: \n#{opt.to_a[1..-1].join("")}
    EOF
    exit
  end

  regdirtree = RegisterDirTree.new(OPTS[:topdir], OPTS[:database], OPTS[:quiet])
  regdirtree.register

end
