desc "Setup all"
task :setup do
  Rake::Task["setup:tables"].invoke
  Rake::Task["setup:tree"].invoke
  Rake::Task["setup:functions"].invoke
end

namespace :setup do
  desc "Setup database tables"
  task :tables => :environment do
    pw = nil
    if User.table_exists? && (root = User.find_by_login('root'))
      print "Do you want to keep root's password? (yes/no, default:yes): "
      res = $stdin.gets.chop!
      unless res == "no"
        pw = root.password
      end
    end
    connection = ActiveRecord::Base.connection
    connection.drop_table("schema_info") if connection.tables.include?("schema_info")
    Rake::Task["db:migrate"].invoke
    Rake::Task["setup:root"].invoke
    if pw
      root = User.find_by_login('root')
      root.password = pw
      root.save!
    end
  end

  desc "Setup tree of directories and variables"
  task :tree => :environment do
    KeywordAttribute.delete_all if KeywordAttribute.table_exists?
    SpatialAttribute.delete_all if SpatialAttribute.table_exists?
    VariableRelation.delete_all if VariableRelation.table_exists?
    Variable.delete_all if Variable.table_exists?
    Directory.delete_all if Directory.table_exists?
    ActualFile.delete_all if ActualFile.table_exists?
    command = 'ruby "' + File.join(File.expand_path(RAILS_ROOT), 'db', 'register_directory_tree.rb') + '"'
    unless system( command )
      raise "failed to execute #{command}"
    end
  end

  desc "Setup functions database"
  task :functions => :environment do
    fixture = File.join(File.expand_path(RAILS_ROOT), 'db', 'function_defaults.yml')
    functions = YAML::load( File.read(fixture) )
    FunctionArgument.delete_all if FunctionArgument.table_exists?
    Type.delete_all if Type.table_exists?
    Type.new(:name => "int").save!
    Type.new(:name => "float").save!
    Type.new(:name => "string").save!
    FunctionOutput.delete_all if FunctionOutput.table_exists?
    Function.delete_all if Function.table_exists?
    functions.each{|k,v|
      arguments = v.delete(:arguments)
      outputs = v.delete(:outputs)
      func = Function.new(v)
      func.save!
      func.function_outputs.create(outputs)
      if arguments
        arguments.each{|arg|
          type = arg.delete(:type)
          fa = FunctionArgument.new(arg)
          fa.type = Type.find_by_name(type)
          fa.function = func
          fa.save!
        }
      end
    }
  end

  desc "Setup root acount"
  task :root => :environment do
    User.delete_all if User.table_exists?
    Group.delete_all if Group.table_exists?
    GroupMember.delete_all if GroupMember.table_exists?
    password = "rootroot"
    user = User.new( :login => "root",
		     :full_name => "default super user",
		     :email_address => GFDNAVI_ADMIN_EMAIL,
		     :affiliation => "GFDNAVI",
                     :password => password,
                     :password_confirmation => password )
    user.super_user = true
    user.save!
  end

end
